using Grasshopper;
using Grasshopper.Kernel;
using Grasshopper.Kernel.Data;
using Rhino;
using Rhino.Geometry;
using Rhino.Geometry.Intersect;
using Rhino.Collections;
using System;
using System.Text;
using System.Collections;
using System.Collections.Generic;
using System.IO;

namespace VesselCam_Project
{
    public class VesselCam_ProjectComponent : GH_Component
    {
       
        /// <summary>
        /// Each implementation of GH_Component must provide a public 
        /// constructor without any arguments.
        /// Category represents the Tab in which the component will appear, 
        /// Subcategory the panel. If you use non-existing tab or panel names, 
        /// new tabs/panels will automatically be created.
        /// </summary>
        public VesselCam_ProjectComponent()
          : base("VesselCam_Project", "Debug1.0",
            "Test Code",
            "Curve", "Primitive")
        {
        }

        /// <summary>
        /// Registers all the input parameters for this component.
        /// </summary>
        protected override void RegisterInputParams(GH_Component.GH_InputParamManager pManager)
        {
            // Use the pManager object to register your input parameters.
            // You can often supply default values when creating parameters.
            // All parameters must have the correct access type. If you want 
            // to import lists or trees of values, modify the ParamAccess flag.
            pManager.AddTextParameter("Author", "author", "Named Person", GH_ParamAccess.item, "Dave");
            pManager.AddBrepParameter("VesselShape", "vesselShape", "Vessel to be Processes", GH_ParamAccess.item);
            pManager.AddBrepParameter("Block", "block", "Block Generic", GH_ParamAccess.item);
            pManager.AddBrepParameter("SliceSurface", "slicePlane", "Plane to Determine Tool Paths", GH_ParamAccess.item);
            pManager.AddTextParameter("CAM Process Type", "camProcessType", "Surface1-2, Rough1-2, Finish1-3", GH_ParamAccess.item);
            pManager.AddNumberParameter("Cutter Diameter", "toolDiam", "Cutter Diameter", GH_ParamAccess.item, 0.125);
            pManager.AddNumberParameter("% Overlap", "pcentOverlap", "Scalar Percent Overlap Per Tool Path", GH_ParamAccess.item, 0.50);
            pManager.AddNumberParameter("Tool Length", "toolLength", "Scalar from Cutter tip to B_Axis Pivot", GH_ParamAccess.item, 2.00);
            pManager.AddIntegerParameter("PtsOnCurve", "ptsOnCrv", "Inner radius for spiral", GH_ParamAccess.item, 10);
            pManager.AddNumberParameter("Cut Depth", "cutDepth", "Cut Depth/Pass", GH_ParamAccess.item, 0.125);
            pManager.AddNumberParameter("Last Cut Depth", "lastCutDepth", "Bottom Cut Depth", GH_ParamAccess.item, 0.0625);
            pManager.AddNumberParameter("Point Simplification Tolerance", "ptReduceToler", "Scalar of Curve Simplification", GH_ParamAccess.item, 0.03);
            pManager.AddBrepParameter("Spindle Shape", "spindleBrep", "Spindle Shape", GH_ParamAccess.item);
            pManager.AddBrepParameter("Cutter Tool Shape", "cutterIntFer", "Shape of Cutter Tool", GH_ParamAccess.item);
            pManager.AddBrepParameter("Machine Shape", "machineIntFer", "Shape of Machine", GH_ParamAccess.item);
            pManager.AddBrepParameter("Workhold Shape", "workholdIntFer", "Shape of WorkHhold", GH_ParamAccess.item);
            // may include a workhold list -> would change to GH_ParamAccess.list
            pManager.AddTextParameter("Job Description", "jobDescipt", "Job Description", GH_ParamAccess.item, "Job Description");
            pManager.AddTextParameter("OutPut Directory", "outDir", "Output NC Directory", GH_ParamAccess.item, "NC output Directory");
            pManager.AddTextParameter("NC  File Name", "ncFileName", "GCode File Name", GH_ParamAccess.item, "GCode NC File name");
            pManager.AddNumberParameter("Safe Surface Clearance", "safeClear", "Safe Clearance Above Surface", GH_ParamAccess.item, 0.01);
            pManager.AddNumberParameter("Skip Tolerance", "skipToler", "Tool Path Skip Tolerance", GH_ParamAccess.item, 0.03);
            pManager.AddTextParameter("Pre Post", "prePost", "Pre Job Setup Blocks", GH_ParamAccess.item, "Pre Job Post");
            pManager.AddTextParameter("Tool Change Post", "toolChangePost", "Tool Change BLocks", GH_ParamAccess.item, "Tool Change Post");
            pManager.AddTextParameter("Shut Down Post", "shutDownPost", "Shut Down Blocks", GH_ParamAccess.item, "ShutDown Block");
            pManager.AddBooleanParameter("Stop?", "stop", "Stop at End of Process", GH_ParamAccess.item, true);
            pManager.AddBooleanParameter("Verbose", "verbose", "Generate a Verbose DataTree for Debugging", GH_ParamAccess.item, true);

            // If you want to change properties of certain parameters, 
            // you can use the pManager instance to access them by index:
            //pManager[0].Optional = true;
        }

        /// <summary>
        /// Registers all the output parameters for this component.
        /// </summary>
        protected override void RegisterOutputParams(GH_Component.GH_OutputParamManager pManager)
        {
            // Use the pManager object to register your output parameters.
            // Output parameters do not have default values, but they too must have the correct access type.
            pManager.AddTextParameter("MessageText", "text", "Generic Text Output", GH_ParamAccess.item);
            pManager.AddTextParameter("Process Text", "processText", "Text Related to This Process", GH_ParamAccess.item);
            pManager.AddTextParameter("GCode Text", "ncText", "NC Text Output", GH_ParamAccess.item);
            pManager.AddGeometryParameter("Subtracted Geometry", "subGeometry", "Block Shape after CNC Process", GH_ParamAccess.list);
            pManager.AddGenericParameter("Debug List", "debugList", "Debug List for Generic Types", GH_ParamAccess.list);
            pManager.AddTextParameter("DebugText", "debugText", "Bebug Text Output", GH_ParamAccess.item);

            // Sometimes you want to hide a specific parameter from the Rhino preview.
            // You can use the HideParameter() method as a quick way:
            //pManager.HideParameter(0);
        }

        /// <summary>
        /// This is the method that actually does the work.
        /// </summary>
        /// <param name="DA">The DA object can be used to retrieve data from input parameters and 
        /// to store data in output parameters.</param>
        protected override void SolveInstance(IGH_DataAccess DA)
        {
            // First, we need to retrieve all data from the input parameters.
            // We'll start by declaring variables and assigning them starting values.
            String author = " ";
            Brep vesselShape = new Brep();
            Brep block = new Brep();
            Brep slicePlane = new Brep();
            string camProcessType = "finish3";
            double toolDiam = 0.125;
            double pcentOverlap = 0.050;
            double toolLength = 2.125;
            int ptsOnCrv = 10;
            double cutDepth = 0.125;
            double lastCutDepth = 0.125;
            double ptReduceToler = 0.125;
            Brep spindleIntFer = new Brep();
            Brep cutterIntFer = new Brep();
            Brep machineIntFer = new Brep();
            Brep workholdIntFer = new Brep();
            String jobDescip = " ";
            String outDir = " ";
            String ncFileName = " ";
            double safeClear = 0.125;
            double skipToler = 0.125;
            String prePost = " ";
            String toolChangePost = " ";
            String shutDownPost = " ";
            bool stop = true;
            bool verbose = true;

            // Then we need to access the input parameters individually. 
            // When data cannot be extracted from a parameter, we should abort this method.
            if (!DA.GetData(0, ref author)) return;
            if (!DA.GetData(1, ref vesselShape)) return;
            if (!DA.GetData(2, ref block)) return;
            if (!DA.GetData(3, ref slicePlane)) return;
            if (!DA.GetData(4, ref camProcessType)) return;
            if (!DA.GetData(5, ref toolDiam)) return;
            if (!DA.GetData(6, ref pcentOverlap)) return;
            if (!DA.GetData(7, ref toolLength)) return;
            if (!DA.GetData(8, ref ptsOnCrv)) return;
            if (!DA.GetData(9, ref cutDepth)) return;
            if (!DA.GetData(10, ref lastCutDepth)) return;
            if (!DA.GetData(11, ref ptReduceToler)) return;
            if (!DA.GetData(12, ref spindleIntFer)) return;
            if (!DA.GetData(13, ref cutterIntFer)) return;
            if (!DA.GetData(14, ref machineIntFer)) return;
            if (!DA.GetData(15, ref workholdIntFer)) return;
            if (!DA.GetData(16, ref jobDescip)) return;
            if (!DA.GetData(17, ref outDir)) return;
            if (!DA.GetData(18, ref ncFileName)) return;
            if (!DA.GetData(19, ref safeClear)) return;
            if (!DA.GetData(20, ref skipToler)) return;
            if (!DA.GetData(21, ref prePost)) return;
            if (!DA.GetData(22, ref toolChangePost)) return;
            if (!DA.GetData(23, ref shutDownPost)) return;
            if (!DA.GetData(24, ref stop)) return;
            if (!DA.GetData(25, ref verbose)) return;

            // We should now validate the data and warn the user if invalid data is supplied.
            if (vesselShape == null)
            {
                AddRuntimeMessage(GH_RuntimeMessageLevel.Error, "Assign a Vessel Shape to Input Port");
                return;
            }
            if (slicePlane == null)
            {
                AddRuntimeMessage(GH_RuntimeMessageLevel.Error, "Assign a Slice Plan to Input Port");
                return;
            }
            if (ptsOnCrv <= 5)
            {
                AddRuntimeMessage(GH_RuntimeMessageLevel.Error, "Not enough Points (< 10)on Slice Curve to Generate GCode");
                return;
            }

            if (toolLength <= 2.00)
            {
                AddRuntimeMessage(GH_RuntimeMessageLevel.Error, "Assign a reasonable tool length.");
                return;
            }

            if (toolDiam < .0625)
            {
                AddRuntimeMessage(GH_RuntimeMessageLevel.Error, "Assign a reasonalble tool diameter.");
                return;
            }

            if (pcentOverlap <= .01)
            {
                AddRuntimeMessage(GH_RuntimeMessageLevel.Error, "Assign a reasonable Overlap Facter in Percent");
                return;
            }

            if (cutterIntFer == null)
            {
                AddRuntimeMessage(GH_RuntimeMessageLevel.Error, "Assign a Cutter Nose Shape to Input Port");
                return;
            }
            if (spindleIntFer == null)
            {
                AddRuntimeMessage(GH_RuntimeMessageLevel.Error, "Assign a Spindle shape to Input Port");
                return;
            }
            // We're set to create the spiral now. To keep the size of the SolveInstance() method small, 
            // The actual functionality will be in a different method:

            // This is the Main Code Area //
            // This is the Main Code Area //
            // This is the Main Code Area //
            // bridge shapes to Custom Additional Code Area



            // duplicate cutter and spindle Breps for orientation to toolLines

#pragma warning disable CS0162 // Unreachable code detected
            Brep cutterIntFerB = new Brep();
#pragma warning restore CS0162 // Unreachable code detected
            cutterIntFerB = makeBrepCopy(cutterIntFer);
            Brep spindleIntFerB = new Brep();
            spindleIntFerB = makeBrepCopy(spindleIntFer);



            // generate slice number based on tool diameter and percent Overlap
            // generate stacking interval for sliceCurves
            // generate degree interval for A_Axis
            double bBoxMaxd = 0.00;
            double rotaRadian = 0.00;
            

            int sliceNumd = calcNumSlices(vesselShape, toolDiam, pcentOverlap, ref bBoxMaxd);
            double rotaSegment = bBoxMaxd / sliceNumd;
            double rotaDegree = calcRotaADegree(sliceNumd);

            // Calculate the cutter tip plane
            Plane planeLoci = SpindleCutterLoci(cutterIntFer, spindleIntFer);

            // rhino calculation tolerance
            double tolerance = 0.0005;
            List<string> gCode = new List<String>();

            DataTree<object> genericList = new DataTree<object>();
            genericList = ProcessGcode(sliceNumd,
                                        rotaRadian,
                                        rotaSegment,
                                        tolerance,
                                        vesselShape,
                                        slicePlane,
                                        ptsOnCrv,
                                        toolLength,
                                        ptReduceToler,
                                        cutterIntFer,
                                        spindleIntFer,
                                        verbose); 

            // This is the Main Code Area //                                        
            // This is the Main Code Area //                                        
            // This is the Main Code Area //
            
            List<Plane> subGeometry = new List<Plane>();  
            DataTree<object> debugList = new DataTree<object>();

            String text = "Text0";
            String processText = "ProcessText1";

            subGeometry.Add(planeLoci);
            debugList = genericList;
            String debugText = "Debug Text5";

            // Finally assign data to the output parameters.
                DA.SetData(0, text);
                DA.SetData(1, processText);
                DA.SetDataList(2, gCode);
                DA.SetDataList(3, subGeometry);
                DA.SetDataTree(4, debugList);
                DA.SetData(5, debugText);


            /// Method Templates
            /// Method Templates
            Brep makeBrepCopy(Brep shape)
            {
                shape.DuplicateBrep();
                return shape;
            }

            int calcNumSlices(Brep _vesselShape, double _toolDiam, double _pcentOverlap, ref double _bBoxMaxd)
            {

                BoundingBox bBox = new BoundingBox();
                Point3d maxPoint = new Point3d();
                Point3d minPoint = new Point3d();
                bBox = _vesselShape.GetBoundingBox(true);
                maxPoint = bBox.Max;
                minPoint = bBox.Min;
                double yPtMax = Math.Abs(maxPoint.Y);
                double zPtMax = Math.Abs(maxPoint.Z);
                double yPtMin = Math.Abs(minPoint.Y);
                double zPtMin = Math.Abs(minPoint.Z);
                double yMax = Math.Max(yPtMax, yPtMin);
                double zMax = Math.Max(zPtMax, zPtMin);
                double maxBox = (Math.PI * 2 * (Math.Max(yMax, zMax)));
                double sliceNum = Math.Round(maxBox / (_toolDiam / (1.00 + _pcentOverlap)));
                int _sliceNumd = (int)Math.Ceiling(sliceNum);
                _bBoxMaxd = maxBox;
                return _sliceNumd;
            }

            double calcRotaADegree(int _sliceNumd)
            {
                rotaRadian = 2 * Math.PI / _sliceNumd;
                double _rotaDegree = RhinoMath.ToDegrees(rotaRadian);
                double _rotaDegree2 = 360.00 / _sliceNumd;
                return _rotaDegree2;
            }


            Plane SpindleCutterLoci(Brep _cutterIntFerB, Brep _spindleIntFerB)
            {
                BoundingBox cBox = new BoundingBox();
                cBox = _cutterIntFerB.GetBoundingBox(true);
                Point3d cmaxPoint = cBox.Max;
                Point3d cminPoint = cBox.Min;

                double cxPtMax = cmaxPoint.X;
                double czPtMax = cmaxPoint.Z;

                double cxPtMin = cminPoint.X;
                double czPtMin = cminPoint.Z;
                double cyPtMin = cminPoint.Y;

                Point3d startLoci = new Point3d((cxPtMax + cxPtMin) / 2 - .5, cyPtMin, (czPtMax + czPtMin) / 2 + .5);
                BoundingBox sBox = new BoundingBox();
                sBox = _spindleIntFerB.GetBoundingBox(true);
                Point3d smaxPoint = sBox.Max;
                Point3d sminPoint = sBox.Min;

                double sxPtMax = smaxPoint.X;
                double szPtMax = smaxPoint.Z;

                double sxPtMin = sminPoint.X;
                double szPtMin = sminPoint.Z;
                double syPtMax = smaxPoint.Y;
                ///TODO
                Point3d endLoci = new Point3d((sxPtMax + sxPtMin) / 2 - .5, syPtMax, (szPtMax + szPtMin) / 2 + .5);
                ///TODO
                Line lineLoci = new Line(startLoci, endLoci);
                Plane _planeLoci = new Plane(startLoci, lineLoci.Direction);
                return _planeLoci;
            }

            Plane tangLine2Loci(Line _tangLine2)
            {
                Plane toolLoci = new Plane(_tangLine2.From, _tangLine2.Direction);
                return toolLoci;
            }


            Transform IntFerRotate(Brep _cutterIntFerB, Brep _spindleIntFerB, Brep _vesselShapeB, Line _tangLine2,
              Line _yAxisLine, double _tolerance)
            {
                Curve[] intFerCurve = new Curve[] { };
                Point3d[] intersectionPoints = new Point3d[] { };
                Transform rota2 = new Transform();
                int j = 0;
                double k;
                // double length;
                // Rotation determine CCW or CW

                if (_tangLine2.FromX <= _yAxisLine.ToX)
                {
                    k = 1.00;
                }
                else
                {
                    k = -1.00;
                }

                for (int i = 1; i < 30; i++)
                {
                    if (Intersection.BrepBrep(_vesselShapeB, _spindleIntFerB, _tolerance, out intFerCurve, out intersectionPoints))
                    {
                        if (null != intersectionPoints)
                        {
                            // Rotate toolLine2 ->
                            rota2 = Transform.Rotation((0.005 * i * k * Math.PI), _tangLine2.From);
                            _tangLine2.Transform(rota2);
                            // Rotate the IntFer Breps - cutter end and Spindle with Shaft
                            _cutterIntFerB.Transform(rota2);
                            _spindleIntFerB.Transform(rota2);
                            j++;

                        }
                        else
                        {
                            goto LoopEnd;
                        }

                    }
                }
                LoopEnd:
                return rota2;
            }



#pragma warning disable CS8321 // Local function is declared but never used
            Transform intFerPullBack(ref List<Brep> _intFerBreps, Brep _vesselShape, Line _tangLine,
#pragma warning restore CS8321 // Local function is declared but never used
              Line _yAxisLine, double _tolerance)
            {
                Curve[] intFerCurve = new Curve[] { };
                Point3d[] intersectionPoints = new Point3d[] { };
                Transform pullBack = new Transform();

                while (Intersection.BrepBrep(_vesselShape, _intFerBreps[0], _tolerance, out intFerCurve, out intersectionPoints))
                {
                    double length = intFerCurve[0].GetLength();
                    if (length > _tolerance)
                    {
                        int j = 1;
                        // pull back toolLine ->
                        Vector3d v0 = _tangLine.Direction;
                        v0.Unitize();
                        pullBack = Transform.Translation(v0 * 0.001f);
                        //tangLine2.Transform(pullBack);
                        // pull back the IntFer Breps - cutter end and Spindle with Shaft
                        _intFerBreps[1].Transform(pullBack);
                        _intFerBreps[0].Transform(pullBack);
                        j++;
                    }
                    else
                    {
                        //String author = " ";
                        break;
                    }

                }
                return pullBack;
            }

            DataTree<object> ProcessGcode(int _sliceNumd,
                                         double _rotaRadian,
                                         double _rotaSegment,
                                         double _tolerance,
                                         Brep _vesselShape,
                                         Brep _slicePlane,
                                         int _ptsOnCurve,
                                         double _toolLength,
                                         double _ptReduceToler,
                                         Brep _cutterinterFer,
                                         Brep _spindleIntFer,
                                         bool _verbose)
            {
                //*****Params used by BrepBrep***                                       
                CurveList curveList = new CurveList();
                Brep tempSlicePlane = slicePlane.DuplicateBrep();

                //Create a List of Breps to hold slicePlane(s)                          
                //Create an array of Point3d to hold ptsOnCurve

                List<Brep> tempSlicePlaneList = new List<Brep>();

                Point3d[] curvePointArray = new Point3d[] { };
                //Line rotaTangline = new Line();
                RhinoList<Point3d> curvePointList = new RhinoList<Point3d>();

                // PolylineCurve polyInterp = new PolylineCurve();
                // Polyline polyInterpLine = new Polyline();
                Curve[] intersectionCurve = new Curve[] { };
                Point3d[] intersectionPoints = new Point3d[] { };
                DataTree<Object> workTree = new DataTree<Object>();
                // Instance Path(0);
                GH_Path sliceCurvePath = new GH_Path(0);          //{0}


                for (int i = 0; i < _sliceNumd; i++)
                {
                    if (tempSlicePlane.Rotate(_rotaRadian, Vector3d.XAxis, Point3d.Origin))
                    {
                        tempSlicePlaneList.Add(tempSlicePlane);
                    }

                    if (Intersection.BrepBrep(_vesselShape, tempSlicePlane, _tolerance, out intersectionCurve, out intersectionPoints))
                    {
                        // Print("Test IntersectionCurve " + intersectionCurve[0].ToString());
                        if (null != intersectionCurve)
                        {

                            intersectionCurve[0].Rotate((-_rotaRadian * i + -_rotaRadian), Vector3d.XAxis, Point3d.Origin);
                            intersectionCurve[0].Translate(0, 0, (i * _rotaSegment));
                            // Print("Rotate Segment" + rotaSegment.ToString());
                            // Reverse Odd interval intersectionCurves
                            if (i % 2 != 0)
                            {
                                intersectionCurve[0].Reverse();
                            }

                            curveList.Add(intersectionCurve[0]);

                        }
                        else
                        {
                            AddRuntimeMessage(GH_RuntimeMessageLevel.Error, "Intersection Failed");

                        }
                    }
                }
                
                        
                // Reverse Curve[_(sliceNumd-1)] at top of the stack
                curveList[_sliceNumd - 1].Reverse();

                // Insert CurveList into Data Tree
                workTree.AddRange(curveList, sliceCurvePath);     //{0}

                int k = 0;// counter for curve params datatree path

                foreach (Curve index in curveList)
                {

                    // Prepare parameters to convert convert curve to PolylineCurve
                    double minimumLength = index.GetLength() / _ptsOnCurve;
                    double maximumLength = minimumLength * 4.00;
                    // Convert the Nurbs Curve to a Polyline
                    PolylineCurve polyInterp = index.ToPolyline(
                      _tolerance,
                      _ptReduceToler,
                      minimumLength,
                      maximumLength);

                    // Collect the number of points of the PolylineCurve
                    int polyInterpPtCount = polyInterp.PointCount;
                    // Store the point counts in a DataTree Path
                    GH_Path slicePtsPath = new GH_Path(1);
                    workTree.Add(polyInterpPtCount, slicePtsPath);    //{1}

                    // Store the PolylineCurve in a DataTree Path
                    GH_Path polyPtsPath = new GH_Path(2);
                    workTree.Add(polyInterp, polyPtsPath);           //{2}

                    // For specific method available to the polyline class
                    // convert a PolylineCurve to Polyline
                    Polyline polyInterpLine = polyInterp.ToPolyline();

                    // get list of curve params for each point on the PolylineCurve
                    // Instance a list to hold actual points on the the Polyline
                    List<Point3d> ptsPolyInterp = new List<Point3d>();
                    // Print("Tolerance  " + _tolerance.ToString());


                    for (int i = 0; i < polyInterpLine.SegmentCount; i++)
                    {
                        // Dump the Polyline Points into the List
                        ptsPolyInterp.Add(polyInterpLine[i]);
                        // Print("polyInterpLine[i]   " + polyInterpLine[i]);
                        // collect the curve parameters at each point on the Polyline, note these are doubles not ints
                        double pointPolyInterpolate = polyInterpLine.ClosestParameter(ptsPolyInterp[i]);
                        // Print("pointPolyInterpolate   " + pointPolyInterpolate);
                        // for each curve parameter collected, calculate the tangent Vector at each point
                        Vector3d tangVector = polyInterp.TangentAt(pointPolyInterpolate);
                        // Correct Vector here

                        // Convert the tangent vector to a line with length of toolLength
                        Line tangLine = new Line(ptsPolyInterp[i], tangVector, _toolLength);

                        // define transforms for rotating the tangLine to a normal from each point
                        Transform rota = Transform.Rotation(-0.5000d * Math.PI, ptsPolyInterp[i]);
                        Transform rotaMirror = Transform.Rotation(-1.5000d * Math.PI, ptsPolyInterp[i]);
                        // test if the the relative polyline is even or odd using k and rotate accordingly
                        //rotate the tangent line to normal

                        if (k % 2 == 0)
                        {
                            tangLine.Transform(rotaMirror);
                        }
                        else
                        {
                            tangLine.Transform(rota);
                        }
                        //Print("TangLine3  " + tangLine.ToString());
                        // Add a reference Vector3d
                        Vector3d yAxisVector = new Vector3d(0.00d, -1.00d, 0.00d);
                        // Create a reference line from the Y Axis

                        Line yAxisLine = new Line(tangLine.To, yAxisVector, _toolLength);
                        // Create a Triangle using the tangent line and the Y Axis reference line
                        Triangle3d angleOffset = new Triangle3d(tangLine.To, yAxisLine.To, tangLine.From);
                        // Calcate the angle between the two lines
                        double reflexAngleRad = RhinoMath.ToDegrees(angleOffset.AngleA);

                        // Determine if the Y Axis line is left or right of the tangent line
                        if (tangLine.FromX <= yAxisLine.ToX)
                        {
                            reflexAngleRad *= -1;
                        }
                        double rotaA_Axis = k * Convert.ToDouble(rotaDegree);


                        // Rotate a Copy of Tangent Lines About the Vessel
                        Line tangLine2 = new Line(tangLine.From, tangLine.To);
                        tangLine2.Transform(Transform.Translation(0, 0, (k * -_rotaSegment)));
                        tangLine2.FromZ = 0;
                        tangLine2.ToZ = 0;
                        tangLine2.Transform(Transform.Rotation((_rotaRadian * k + _rotaRadian), Vector3d.XAxis, Point3d.Origin));

                        Plane cutnosePlane = SpindleCutterLoci(cutterIntFer, _spindleIntFer);
                        Plane tangLine2Plane = tangLine2Loci(tangLine2);

                        bool cutterIntFerBool = cutterIntFer.Transform(Transform.ChangeBasis(cutnosePlane, tangLine2Plane));
                        if (!cutterIntFerBool)
                        {
                            AddRuntimeMessage(GH_RuntimeMessageLevel.Error, "Cutter transform a success");

                        }

                        bool spindleIntFerBool = _spindleIntFer.Transform(Transform.ChangeBasis(cutnosePlane, tangLine2Plane));
                        if (!spindleIntFerBool)
                        {
                            AddRuntimeMessage(GH_RuntimeMessageLevel.Error, "Spindle transform a success");

                        }


                        //Print("cutnosePlane   " + cutnosePlane.Origin.ToString());

                        //  Test the interference with the cutter shafte and spindle and the vessel and adjust toolLine2
                        Transform xAngle = new Transform();
                        xAngle = IntFerRotate(cutterIntFer, _spindleIntFer, _vesselShape, tangLine2, yAxisLine, _tolerance);
                        // Print("xangle  " + xAngle.ToString());
                        tangLine2.Transform(xAngle);

                        // Test the interference with the cutter head and the vessel and adjust toolLine2
                        // Transform xPullBack = new Transform();
                        // xPullBack = intFerPullBack(_cutterIntFer, _spindleIntFer, _vesselShape, tangLine2, yAxisLine, _tolerance, _planeLoci);
                        // tangLine2.Transform(xPullBack);
                        // Calculate a new B_Axis Angle for tangLing2



                        Line yAxisLine2 = new Line(tangLine2.To, yAxisVector, _toolLength);
                        Triangle3d angleOffset2 = new Triangle3d(tangLine2.To, yAxisLine2.To, tangLine2.From);

                        double reflexAngleRad2 = RhinoMath.ToDegrees(angleOffset2.AngleA);

                        // Determine if the Y Axis line is left or right of the tangent line
                        if (tangLine2.FromX <= yAxisLine.ToX)
                        {
                            reflexAngleRad2 *= -1;
                        }

                        Line tangLineRadial = tangLine2;

                        String gCodeBlock = String.Format(
                            "G01 X" + tangLine.FromX.ToString("00.0000") +
                            ", Y" + tangLine.FromY.ToString("00.0000") +
                            ", Z00.000, A" + rotaA_Axis.ToString("000.000") +
                            ", B" + reflexAngleRad.ToString("00.0000"));

                        gCode.Add(gCodeBlock);

                        if (_verbose) // use to debug
                        {
                            GH_Path ptsSlicePath = new GH_Path(3, k);
                            GH_Path sliceParamPath = new GH_Path(4, k);
                            GH_Path tangVectorPath = new GH_Path(5, k);
                            GH_Path tangLinePath = new GH_Path(6, k);
                            GH_Path tangNormPath = new GH_Path(7, k);
                            GH_Path yAxisPath = new GH_Path(8, k);
                            GH_Path reflexBAxisPath = new GH_Path(9, k);
                            GH_Path ptsXBAxisPath = new GH_Path(10, k);
                            GH_Path ptsYBAxisPath = new GH_Path(11, k);
                            GH_Path ptsZBAxisPath = new GH_Path(12, k);
                            GH_Path reflexAAxisPath = new GH_Path(13);
                            GH_Path tangNormalRadialPath = new GH_Path(14, k);
                            GH_Path reflexBAxisPath2 = new GH_Path(15, k);
                            GH_Path spindleBrepPath = new GH_Path(16, k);
                            // Dump Polyline Points into the DataTree
                            workTree.Add(polyInterpLine[i], ptsSlicePath);      //{3, k}
                                                                                // Dump the each parameter into a DataTree with a looping index
                            workTree.Add(pointPolyInterpolate, sliceParamPath); //{4, K}
                                                                                // Dump the tangent Vector and tang Line into the DataTree
                            workTree.Add(tangVector, tangVectorPath);           //{5, k}
                            workTree.Add(tangLine, tangLinePath);               //{6 ,k}
                                                                                // Dump the tangent line into the DataTree
                            workTree.Add(tangLine, tangNormPath);               //{7, k}
                                                                                // Dump the Y Axis reference line into the DataTree
                            workTree.Add(yAxisLine, yAxisPath);                 //{8,k}
                                                                                // Dump the reflex angle in to the DataTree
                            workTree.Add(reflexAngleRad, reflexBAxisPath);      //{9, k}
                                                                                // Get and Dump the X value of the B Axis pivot point
                            workTree.Add(tangLine.FromX, ptsXBAxisPath);        //{10,k}
                                                                                // Get and Dump the y value of the B Axis pivot point
                            workTree.Add(tangLine.FromY, ptsYBAxisPath);        //{11,k}
                                                                                // Get and Dump the z value of the B Axis pivot point
                            workTree.Add(tangLine.FromZ, ptsZBAxisPath);        //{12,k}
                                                                                // Get and Dump A_Axis Rotation Angle
                            workTree.Add(rotaA_Axis, reflexAAxisPath);          //{13}
                            workTree.Add(tangLineRadial, tangNormalRadialPath); // (14,k)
                            workTree.Add(reflexAngleRad, reflexBAxisPath2);     // (15,k)
                            workTree.Add(_spindleIntFer, spindleBrepPath);      // (16,k)

                        }


                    }
                    k++;

                }// curve points process

                // Dump the rotation shift A Axis


                // each sliceCurve processing


                return workTree;
            }
            
        }  // End of SolveInstance Method
        
        /// <summary>
        /// The Exposure property controls where in the panel a component icon 
        /// will appear. There are seven possible locations (primary to septenary), 
        /// each of which can be combined with the GH_Exposure.obscure flag, which 
        /// ensures the component will only be visible on panel dropdowns.
        /// </summary>
        public override GH_Exposure Exposure => GH_Exposure.primary;

        /// <summary>
        /// Provides an Icon for every component that will be visible in the User Interface.
        /// Icons need to be 24x24 pixels.
        /// You can add image files to your project resources and access them like this:
        /// return Resources.IconForThisComponent;
        /// </summary>
        protected override System.Drawing.Bitmap Icon => null;

        /// <summary>
        /// Each component must have a unique Guid to identify it. 
        /// It is vital this Guid doesn't change otherwise old ghx files 
        /// that use the old ID will partially fail during loading.
        /// </summary>
        public override Guid ComponentGuid => new Guid("4347D264-87E8-4781-BDAA-D98FCEA79541");
    }
}